@@ -30,11 +30,12 @@
 #include <polkit/polkit.h>
 
 #include "cc-wifi-hotspot-dialog.h"
-#include "list-box-helper.h"
+#include "cc-hostname.h"
 #include "hostname-helper.h"
 #include "network-dialogs.h"
 #include "panel-common.h"
 #include "cc-list-row.h"
+#include "cc-qr-code-dialog.h"
 
 #include "connection-editor/net-connection-editor.h"
 #include "net-device-wifi.h"
@@ -47,24 +48,35 @@
 static void nm_device_wifi_refresh_ui (NetDeviceWifi *self);
 static void show_wifi_list (NetDeviceWifi *self);
 static void show_hotspot_ui (NetDeviceWifi *self);
+static void nm_client_on_permission_change (NetDeviceWifi *self);
 
 
 struct _NetDeviceWifi
 {
-        GtkStack                 parent;
+        AdwBin                   parent;
 
-        GtkBox                  *center_box;
-        GtkButton               *connect_hidden_button;
+        AdwWindowTitle          *wifi_headerbar_title;
+        AdwPreferencesGroup     *details_box;
         GtkSwitch               *device_off_switch;
         GtkBox                  *header_box;
+        GtkPopover              *header_button_popover;
         GtkBox                  *hotspot_box;
         CcListRow               *hotspot_name_row;
         CcListRow               *hotspot_security_row;
         CcListRow               *hotspot_password_row;
         GtkBox                  *listbox_box;
-        GtkButton               *start_hotspot_button;
-        GtkLabel                *status_label;
-        GtkLabel                *title_label;
+        GtkStack                *stack;
+        AdwPreferencesGroup     *saved_networks_box;
+        CcWifiConnectionList    *saved_networks_list;
+        GtkWindow               *saved_networks_dialog;
+        AdwToastOverlay         *saved_networks_toast_overlay;
+        AdwToast                *saved_networks_undo_toast;
+        GPtrArray               *saved_networks_forgotten_rows;
+
+        AdwSwitchRow            *device_enable_row;
+        CcListRow               *saved_network_row;
+        CcListRow               *connect_hidden_row;
+        CcListRow               *hotspot_row;
 
         CcPanel                 *panel;
         NMClient                *client;
@@ -89,20 +101,14 @@ enum {
         PROP_LAST,
 };
 
-G_DEFINE_TYPE (NetDeviceWifi, net_device_wifi, GTK_TYPE_STACK)
+G_DEFINE_TYPE (NetDeviceWifi, net_device_wifi, ADW_TYPE_BIN)
 
 static void
 disable_scan_timeout (NetDeviceWifi *self)
 {
         g_debug ("Disabling periodic Wi-Fi scan");
-        if (self->monitor_scanning_id > 0) {
-                g_source_remove (self->monitor_scanning_id);
-                self->monitor_scanning_id = 0;
-        }
-        if (self->scan_id > 0) {
-                g_source_remove (self->scan_id);
-                self->scan_id = 0;
-        }
+        g_clear_handle_id (&self->monitor_scanning_id, g_source_remove);
+        g_clear_handle_id (&self->scan_id, g_source_remove);
 }
 
 static void
@@ -113,10 +119,14 @@ wireless_enabled_toggled (NetDeviceWifi *self)
         enabled = nm_client_wireless_get_enabled (self->client);
 
         self->updating_device = TRUE;
-        gtk_switch_set_active (self->device_off_switch, enabled);
-        if (!enabled)
+        g_object_set (self->device_enable_row, "active", enabled, NULL);
+        if (!enabled) {
                 disable_scan_timeout (self);
-        gtk_widget_set_sensitive (GTK_WIDGET (self->connect_hidden_button), enabled);
+                gtk_widget_set_sensitive (GTK_WIDGET (self->hotspot_row), FALSE);
+        } else {
+                nm_client_on_permission_change(self);
+        }
+        gtk_widget_set_sensitive (GTK_WIDGET (self->connect_hidden_row), enabled);
         self->updating_device = FALSE;
 }
 
@@ -218,8 +228,7 @@ device_get_hotspot_security_details (NetDeviceWifi *self,
         tmp_security = C_("Wifi security", "None");
 
         /* Key management values:
-         * "none" = WEP
-         * "wpa-none" = WPAv1 Ad-Hoc mode (not supported in NM >= 0.9.4)
+         * "none" = WEP or no password protection
          * "wpa-psk" = WPAv2 Ad-Hoc mode (eg IBSS RSN) and AP-mode WPA v1 and v2
          */
         key_mgmt = nm_setting_wireless_security_get_key_mgmt (sws);
@@ -227,8 +236,7 @@ device_get_hotspot_security_details (NetDeviceWifi *self,
                 tmp_secret = nm_setting_wireless_security_get_wep_key (sws, 0);
                 tmp_security = _("WEP");
         }
-        else if (strcmp (key_mgmt, "wpa-none") == 0 ||
-                 strcmp (key_mgmt, "wpa-psk") == 0) {
+        else if (strcmp (key_mgmt, "wpa-psk") == 0) {
                 tmp_secret = nm_setting_wireless_security_get_psk (sws);
                 tmp_security = _("WPA");
         } else {
@@ -361,10 +369,11 @@ nm_device_wifi_refresh_ui (NetDeviceWifi *self)
         wireless_enabled_toggled (self);
 
         status = panel_device_status_to_localized_string (self->device, NULL);
-        gtk_label_set_label (self->status_label, status);
-
+        adw_window_title_set_subtitle (self->wifi_headerbar_title, status);
         /* update list of APs */
         show_wifi_list (self);
+
+        gtk_widget_set_visible (GTK_WIDGET (self->saved_network_row), !cc_wifi_connection_list_is_empty (self->saved_networks_list));
 }
 
 static void
@@ -375,20 +384,24 @@ device_off_switch_changed_cb (NetDeviceWifi *self)
         if (self->updating_device)
                 return;
 
-        active = gtk_switch_get_active (self->device_off_switch);
-        nm_client_wireless_set_enabled (self->client, active);
+        active = adw_switch_row_get_active (self->device_enable_row);
+        nm_client_dbus_set_property (self->client,
+                                     NM_DBUS_PATH,
+                                     NM_DBUS_INTERFACE,
+                                     "WirelessEnabled",
+                                     g_variant_new_boolean (active),
+                                     -1,
+                                     NULL, NULL, NULL);
         if (!active)
                 disable_scan_timeout (self);
-        gtk_widget_set_sensitive (GTK_WIDGET (self->connect_hidden_button), active);
+        gtk_widget_set_sensitive (GTK_WIDGET (self->connect_hidden_row), active);
 }
 
 static void
-connect_hidden_button_clicked_cb (NetDeviceWifi *self)
+connect_hidden (NetDeviceWifi *self)
 {
-        GtkWidget *toplevel;
-
-        toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
-        cc_network_panel_connect_to_hidden_network (toplevel, self->client);
+        GtkNative *native = gtk_widget_get_native (GTK_WIDGET (self));
+        cc_network_panel_connect_to_hidden_network (GTK_WIDGET (native), self->client);
 }
 
 static void
@@ -513,42 +526,6 @@ wireless_try_to_connect (NetDeviceWifi *self,
         }
 }
 
-static gchar *
-get_hostname (void)
-{
-        g_autoptr(GDBusConnection) bus = NULL;
-        g_autoptr(GVariant) res = NULL;
-        g_autoptr(GVariant) inner = NULL;
-        g_autoptr(GError) error = NULL;
-
-        bus = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
-        if (bus == NULL) {
-                g_warning ("Failed to get system bus connection: %s", error->message);
-                return NULL;
-        }
-        res = g_dbus_connection_call_sync (bus,
-                                           "org.freedesktop.hostname1",
-                                           "/org/freedesktop/hostname1",
-                                           "org.freedesktop.DBus.Properties",
-                                           "Get",
-                                           g_variant_new ("(ss)",
-                                                          "org.freedesktop.hostname1",
-                                                          "PrettyHostname"),
-                                           (GVariantType*)"(v)",
-                                           G_DBUS_CALL_FLAGS_NONE,
-                                           -1,
-                                           NULL,
-                                           &error);
-
-        if (res == NULL) {
-                g_warning ("Getting pretty hostname failed: %s", error->message);
-                return NULL;
-        }
-
-        g_variant_get (res, "(v)", &inner);
-        return g_variant_dup_string (inner, NULL);
-}
-
 static gboolean
 is_hotspot_connection (NMConnection *connection)
 {
@@ -582,7 +559,7 @@ static void
 show_hotspot_ui (NetDeviceWifi *self)
 {
         /* show hotspot tab */
-        gtk_stack_set_visible_child (GTK_STACK (self), GTK_WIDGET (self->hotspot_box));
+        gtk_stack_set_visible_child (self->stack, GTK_WIDGET (self->hotspot_box));
 }
 
 static void
@@ -676,29 +653,10 @@ overwrite_ssid_cb (GObject      *source_object,
 }
 
 static void
-start_hotspot_button_clicked_cb (NetDeviceWifi *self)
+on_wifi_hotspot_dialog_response_cb (GtkDialog     *dialog,
+                                   gint           response,
+                                   NetDeviceWifi *self)
 {
-        g_autofree gchar *active_ssid = NULL;
-        GtkWidget *window;
-        NMConnection *c;
-        g_autofree gchar *hostname = NULL;
-        g_autofree gchar *ssid = NULL;
-        gint response;
-
-        window = gtk_widget_get_toplevel (GTK_WIDGET (self));
-
-        if (!self->hotspot_dialog)
-                self->hotspot_dialog = cc_wifi_hotspot_dialog_new (GTK_WINDOW (window));
-        cc_wifi_hotspot_dialog_set_device (self->hotspot_dialog, NM_DEVICE_WIFI (self->device));
-        hostname = get_hostname ();
-        ssid =  pretty_hostname_to_ssid (hostname);
-        cc_wifi_hotspot_dialog_set_hostname (self->hotspot_dialog, ssid);
-                c = net_device_wifi_get_hotspot_connection (self);
-        if (c)
-                cc_wifi_hotspot_dialog_set_connection (self->hotspot_dialog, c);
-
-        response = gtk_dialog_run (GTK_DIALOG (self->hotspot_dialog));
-
         if (response == GTK_RESPONSE_APPLY) {
                 NMConnection *connection;
 
@@ -719,7 +677,33 @@ start_hotspot_button_clicked_cb (NetDeviceWifi *self)
                                                                      self);
         }
 
-        gtk_widget_hide (GTK_WIDGET (self->hotspot_dialog));
+        gtk_widget_set_visible (GTK_WIDGET (self->hotspot_dialog), FALSE);
+}
+
+static void
+start_hotspot (NetDeviceWifi *self)
+{
+        GtkNative *native;
+        NMConnection *c;
+        g_autofree gchar *hostname = NULL;
+        g_autofree gchar *ssid = NULL;
+
+        native = gtk_widget_get_native (GTK_WIDGET (self));
+
+        if (!self->hotspot_dialog) {
+                self->hotspot_dialog = cc_wifi_hotspot_dialog_new (GTK_WINDOW (native));
+                g_object_ref_sink (self->hotspot_dialog);
+        }
+        cc_wifi_hotspot_dialog_set_device (self->hotspot_dialog, NM_DEVICE_WIFI (self->device));
+        hostname = cc_hostname_get_display_hostname (cc_hostname_get_default ());
+        ssid =  pretty_hostname_to_ssid (hostname);
+        cc_wifi_hotspot_dialog_set_hostname (self->hotspot_dialog, ssid);
+                c = net_device_wifi_get_hotspot_connection (self);
+        if (c)
+                cc_wifi_hotspot_dialog_set_connection (self->hotspot_dialog, c);
+
+        g_signal_connect_after (self->hotspot_dialog, "response", G_CALLBACK (on_wifi_hotspot_dialog_response_cb), self);
+        gtk_window_present (GTK_WINDOW (self->hotspot_dialog));
 }
 
 static void
@@ -737,7 +721,7 @@ stop_shared_connection (NetDeviceWifi *self)
 
                 devices = nm_active_connection_get_devices (c);
                 if (devices && devices->pdata[0] == self->device) {
-                        nm_client_deactivate_connection (self->client, c, NULL, NULL);
+                        nm_client_deactivate_connection_async (self->client, c, NULL, NULL, NULL);
                         found = TRUE;
                         break;
                 }
@@ -754,7 +738,23 @@ stop_shared_connection (NetDeviceWifi *self)
 static void
 show_wifi_list (NetDeviceWifi *self)
 {
-        gtk_stack_set_visible_child (GTK_STACK (self), GTK_WIDGET (self->listbox_box));
+        gtk_stack_set_visible_child (self->stack, GTK_WIDGET (self->listbox_box));
+}
+
+static void
+net_device_wifi_dispose (GObject *object)
+{
+        NetDeviceWifi *self = NET_DEVICE_WIFI (object);
+
+        if (self->hotspot_dialog) {
+                gtk_window_destroy (GTK_WINDOW (self->hotspot_dialog));
+                g_object_unref (self->hotspot_dialog);
+                self->hotspot_dialog = NULL;
+        }
+
+        g_clear_pointer (&self->saved_networks_dialog, gtk_window_destroy);
+
+        G_OBJECT_CLASS (net_device_wifi_parent_class)->dispose (object);
 }
 
 static void
@@ -768,6 +768,7 @@ net_device_wifi_finalize (GObject *object)
 
         g_clear_object (&self->client);
         g_clear_object (&self->device);
+        g_clear_object (&self->saved_networks_list);
         g_clear_pointer (&self->selected_ssid_title, g_free);
         g_clear_pointer (&self->selected_connection_id, g_free);
         g_clear_pointer (&self->selected_ap_id, g_free);
@@ -798,88 +799,104 @@ really_forgotten (GObject              *source_object,
                   GAsyncResult         *res,
                   gpointer              user_data)
 {
-        g_autoptr(CcWifiConnectionList) list = user_data;
+        g_autoptr(NetDeviceWifi) self = NET_DEVICE_WIFI (user_data);
         g_autoptr(GError) error = NULL;
 
-        cc_wifi_connection_list_thaw (list);
+        cc_wifi_connection_list_thaw (self->saved_networks_list);
 
         if (!nm_remote_connection_delete_finish (NM_REMOTE_CONNECTION (source_object), res, &error))
                 g_warning ("failed to delete connection %s: %s",
                            nm_object_get_path (NM_OBJECT (source_object)),
                            error->message);
+
+        if (cc_wifi_connection_list_is_empty (self->saved_networks_list)) {
+            gtk_window_close (self->saved_networks_dialog);
+            gtk_widget_set_visible (GTK_WIDGET (self->saved_network_row), FALSE);
+        }
 }
 
 static void
-really_forget (GtkDialog *dialog, gint response, gpointer data)
+really_forget (gpointer data)
 {
-        GtkWidget *forget = data;
-        CcWifiConnectionRow *row;
-        GList *rows;
-        GList *r;
+        CcWifiConnectionRow *row = CC_WIFI_CONNECTION_ROW (data);
         NMRemoteConnection *connection;
         NetDeviceWifi *self;
 
-        gtk_widget_destroy (GTK_WIDGET (dialog));
+        self = NET_DEVICE_WIFI (g_object_get_data (G_OBJECT (row), "net"));
 
-        if (response != GTK_RESPONSE_OK)
-                return;
+        connection = NM_REMOTE_CONNECTION (cc_wifi_connection_row_get_connection (row));
 
-        self = NET_DEVICE_WIFI (g_object_get_data (G_OBJECT (forget), "net"));
-        rows = g_object_steal_data (G_OBJECT (forget), "rows");
-        for (r = rows; r; r = r->next) {
-                row = r->data;
-                connection = NM_REMOTE_CONNECTION (cc_wifi_connection_row_get_connection (row));
-                nm_remote_connection_delete_async (connection, self->cancellable, really_forgotten, g_object_ref (data));
-                gtk_widget_hide (GTK_WIDGET (row));
-        }
-        g_list_free (rows);
+        nm_remote_connection_delete_async (connection, self->cancellable, really_forgotten, g_object_ref (self));
 }
 
 static void
-forget_selected (GtkButton *forget, CcWifiConnectionList *list)
+undo_forget (gpointer data)
 {
-        GtkWidget *window;
-        GtkWidget *dialog;
+        CcWifiConnectionRow *row = CC_WIFI_CONNECTION_ROW (data);
+        NetDeviceWifi *self;
 
-        window = gtk_widget_get_toplevel (GTK_WIDGET (forget));
-        dialog = gtk_message_dialog_new (GTK_WINDOW (window),
-                                         GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                         GTK_MESSAGE_OTHER,
-                                         GTK_BUTTONS_NONE,
-                                         NULL);
-        gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG (dialog),
-                                       _("Network details for the selected networks, including passwords and any custom configuration will be lost."));
+        self = NET_DEVICE_WIFI (g_object_get_data (G_OBJECT (row), "net"));
 
-        gtk_dialog_add_buttons (GTK_DIALOG (dialog),
-                                _("_Cancel"), GTK_RESPONSE_CANCEL,
-                                _("_Forget"), GTK_RESPONSE_OK,
-                                NULL);
-        g_signal_connect (dialog, "response",
-                          G_CALLBACK (really_forget), list);
-        gtk_window_present (GTK_WINDOW (dialog));
+        gtk_widget_set_visible (GTK_WIDGET (row), TRUE);
+
+        cc_wifi_connection_list_thaw (self->saved_networks_list);
 }
 
 static void
-check_toggled (CcWifiConnectionRow *row, GParamSpec *pspec, CcWifiConnectionList *list)
+on_saved_networks_forget_dismissed (NetDeviceWifi *self)
 {
-        GtkWidget *forget;
-        gboolean active;
-        GList *rows;
+        self->saved_networks_undo_toast = NULL;
 
-        active = cc_wifi_connection_row_get_checked (row);
-        rows = g_object_steal_data (G_OBJECT (list), "rows");
+        /* Since the free function is set to either undo or really forget,
+           this always does the right thing */
+        g_clear_pointer (&self->saved_networks_forgotten_rows, g_ptr_array_unref);
+}
 
-        if (active) {
-                cc_wifi_connection_list_freeze (list);
-                rows = g_list_prepend (rows, row);
+static void
+on_saved_networks_forget_undo (NetDeviceWifi *self)
+{
+        g_ptr_array_set_free_func (self->saved_networks_forgotten_rows, undo_forget);
+}
+
+static void
+forget_selected (NetDeviceWifi *self, CcWifiConnectionRow *row, CcWifiConnectionList *list)
+{
+        g_autofree gchar *message = NULL;
+
+        cc_wifi_connection_list_freeze (list);
+
+        gtk_widget_set_visible (GTK_WIDGET (row), FALSE);
+        g_object_set_data (G_OBJECT (row), "net", self);
+
+        /* Add row to the array of rows to be handled by the undo toast */
+        if (!self->saved_networks_forgotten_rows)
+                self->saved_networks_forgotten_rows = g_ptr_array_new_with_free_func (really_forget);
+
+        g_ptr_array_add (self->saved_networks_forgotten_rows, row);
+
+        /* If undo toast does not exist, create it, otherwise update title and extend timeout */
+        if (!self->saved_networks_undo_toast) {
+                message = g_strdup (_("Network deleted"));
+                self->saved_networks_undo_toast = adw_toast_new (message);
+                adw_toast_set_button_label (self->saved_networks_undo_toast, _("_Undo"));
+
+                g_signal_connect_swapped (self->saved_networks_undo_toast,
+                                          "button-clicked",
+                                          G_CALLBACK (on_saved_networks_forget_undo),
+                                          self);
+
+                g_signal_connect_swapped (self->saved_networks_undo_toast,
+                                          "dismissed",
+                                          G_CALLBACK (on_saved_networks_forget_dismissed),
+                                          self);
         } else {
-                cc_wifi_connection_list_thaw (list);
-                rows = g_list_remove (rows, row);
+                message = g_strdup_printf (_("%d networks deleted"), self->saved_networks_forgotten_rows->len);
+                adw_toast_set_title (self->saved_networks_undo_toast, message);
+
+                g_object_ref (self->saved_networks_undo_toast);
         }
 
-        g_object_set_data_full (G_OBJECT (list), "rows", rows, (GDestroyNotify)g_list_free);
-        forget = g_object_get_data (G_OBJECT (list), "forget");
-        gtk_widget_set_sensitive (forget, rows != NULL);
+        adw_toast_overlay_add_toast (self->saved_networks_toast_overlay, self->saved_networks_undo_toast);
 }
 
 static gint
@@ -962,15 +979,6 @@ ap_sort (gconstpointer a, gconstpointer b, gpointer data)
         return 0;
 }
 
-static gboolean
-on_focus_out_event (GtkWindow *window,
-                    GdkEvent  *event)
-{
-  gtk_window_close(window);  
-
-  return TRUE;
-}
-
 static void
 show_details_for_row (NetDeviceWifi *self, CcWifiConnectionRow *row, CcWifiConnectionList *list)
 {
@@ -982,159 +990,48 @@ show_details_for_row (NetDeviceWifi *self, CcWifiConnectionRow *row, CcWifiConne
         ap = cc_wifi_connection_row_best_access_point (row);
 
         editor = net_connection_editor_new (connection, self->device, ap, self->client);
-        g_signal_connect (GTK_WINDOW (editor),
-                    "focus-out-event",
-                    G_CALLBACK(on_focus_out_event),
-                    GTK_WINDOW (editor));  
-        gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (row))));
+        gtk_window_set_transient_for (GTK_WINDOW (editor), GTK_WINDOW (gtk_widget_get_native (GTK_WIDGET (row))));
         gtk_window_present (GTK_WINDOW (editor));
 }
 
 static void
-on_connection_list_row_added_cb (NetDeviceWifi        *self,
-                                 CcWifiConnectionRow  *row,
-                                 CcWifiConnectionList *list)
-{
-        g_signal_connect (row, "notify::checked",
-                          G_CALLBACK (check_toggled), list);
-}
-
-static void
-on_connection_list_row_removed_cb (NetDeviceWifi        *self,
-                                   CcWifiConnectionRow  *row,
-                                   CcWifiConnectionList *list)
-{
-        GList *rows = g_object_steal_data (G_OBJECT (list), "rows");
-        GList *item;
-        GtkWidget *forget;
-
-        item = g_list_find (rows, row);
-        if (item)
-          {
-            rows = g_list_remove (rows, row);
-            cc_wifi_connection_list_thaw (list);
-          }
-        g_object_set_data_full (G_OBJECT (list), "rows", rows, (GDestroyNotify)g_list_free);
-
-        forget = g_object_get_data (G_OBJECT (list), "forget");
-        gtk_widget_set_sensitive (forget, rows != NULL);
-}
-
-static void
-on_connection_list_row_activated_cb (NetDeviceWifi        *self,
-                                     CcWifiConnectionRow  *row,
-                                     CcWifiConnectionList *list)
-{
-  cc_wifi_connection_row_set_checked (row, !cc_wifi_connection_row_get_checked (row));
-}
-
-static void
-history_button_clicked_cb (NetDeviceWifi *self)
+show_qr_code_for_row (NetDeviceWifi *self, CcWifiConnectionRow *row, CcWifiConnectionList *list)
 {
+        NMConnection *connection;
+        GtkNative *native;
         GtkWidget *dialog;
-        GtkWidget *window;
-        GtkWidget *forget;
-        GtkWidget *header;
-        GtkWidget *swin;
-        GtkWidget *content_area;
-        GtkWidget *separator;
-        GtkWidget *list;
-        GList *list_rows;
 
-        dialog = g_object_new (GTK_TYPE_DIALOG, "use-header-bar", 1, NULL);
-        window = gtk_widget_get_toplevel (GTK_WIDGET (self));
-        gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
-        g_signal_connect (GTK_WINDOW (dialog),
-                    "focus-out-event",
-                    G_CALLBACK(on_focus_out_event),
-                    GTK_WINDOW (dialog));
-        gtk_window_set_title (GTK_WINDOW (dialog), _("Known Wi-Fi Networks"));
-        gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
-        gtk_window_set_default_size (GTK_WINDOW (dialog), 500, 400);
+        connection = cc_wifi_connection_row_get_connection (row);
 
-        /* Dialog's header */
-        header = gtk_header_bar_new ();
-        gtk_widget_show (header);
-        gtk_header_bar_set_title (GTK_HEADER_BAR (header), _("Known Wi-Fi Networks"));
-        gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (header), TRUE);
-        gtk_window_set_titlebar (GTK_WINDOW (dialog), header);
-
-        g_signal_connect (dialog, "response",
-                          G_CALLBACK (gtk_widget_destroy), NULL);
-
-        g_signal_connect (dialog, "delete-event",
-                          G_CALLBACK (gtk_widget_destroy), NULL);
-
-        content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
-        gtk_container_set_border_width (GTK_CONTAINER (content_area), 0);
-
-        swin = gtk_scrolled_window_new (NULL, NULL);
-        gtk_widget_show (swin);
-        gtk_widget_set_vexpand (swin, TRUE);
-        gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swin), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
-        gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (swin), GTK_SHADOW_NONE);
-        gtk_container_add (GTK_CONTAINER (content_area), swin);
-
-        list = GTK_WIDGET (cc_wifi_connection_list_new (self->client,
-                                                        NM_DEVICE_WIFI (self->device),
-                                                        FALSE, FALSE, TRUE));
-        gtk_widget_show (list);
-        gtk_list_box_set_selection_mode (GTK_LIST_BOX (list), GTK_SELECTION_NONE);
-        gtk_list_box_set_header_func (GTK_LIST_BOX (list), cc_list_box_update_header_func, NULL, NULL);
-        gtk_list_box_set_sort_func (GTK_LIST_BOX (list), (GtkListBoxSortFunc)history_sort, NULL, NULL);
-        gtk_container_add (GTK_CONTAINER (swin), list);
-
-        /* Horizontal separator */
-        separator = gtk_separator_new (GTK_ORIENTATION_HORIZONTAL);
-        gtk_widget_show (separator);
-        gtk_container_add (GTK_CONTAINER (content_area), separator);
-
-        /* translators: This is the label for the "Forget wireless network" functionality */
-        forget = gtk_button_new_with_mnemonic (C_("Wi-Fi Network", "_Forget"));
-        gtk_widget_show (forget);
-        gtk_widget_set_halign (forget, GTK_ALIGN_START);
-        gtk_widget_set_margin_top (forget, 6);
-        gtk_widget_set_margin_bottom (forget, 6);
-        gtk_widget_set_margin_start (forget, 6);
-        gtk_widget_set_sensitive (forget, FALSE);
-        gtk_style_context_add_class (gtk_widget_get_style_context (forget), "destructive-action");
-
-        g_signal_connect (forget, "clicked",
-                          G_CALLBACK (forget_selected), list);
-        gtk_container_add (GTK_CONTAINER (content_area), forget);
-
-        g_object_set_data (G_OBJECT (list), "forget", forget);
-        g_object_set_data (G_OBJECT (list), "net", self);
-
-        g_signal_connect_object (list, "add",
-                                 G_CALLBACK (on_connection_list_row_added_cb),
-                                 self,
-                                 G_CONNECT_SWAPPED);
-        g_signal_connect_object (list, "remove",
-                                 G_CALLBACK (on_connection_list_row_removed_cb),
-                                 self,
-                                 G_CONNECT_SWAPPED);
-        g_signal_connect_object (list, "row-activated",
-                                 G_CALLBACK (on_connection_list_row_activated_cb),
-                                 self,
-                                 G_CONNECT_SWAPPED);
-        g_signal_connect_object (list, "configure",
-                                 G_CALLBACK (show_details_for_row),
-                                 self,
-                                 G_CONNECT_SWAPPED);
-
-        list_rows = gtk_container_get_children (GTK_CONTAINER (list));
-        while (list_rows)
-          {
-            on_connection_list_row_added_cb (self,
-                                             CC_WIFI_CONNECTION_ROW (list_rows->data),
-                                             CC_WIFI_CONNECTION_LIST (list));
-            list_rows = g_list_delete_link (list_rows, list_rows);
-          }
+        // getting a new "local" connection, since we don't want to populate the secrets of the original connection
+        connection = NM_CONNECTION (nm_client_get_connection_by_id (self->client, nm_connection_get_id (connection)));
 
+        dialog = cc_qr_code_dialog_new (connection);
+        native = gtk_widget_get_native (GTK_WIDGET (self));
+        gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (native));
         gtk_window_present (GTK_WINDOW (dialog));
 }
 
+static void
+on_show_saved_network (NetDeviceWifi *self)
+{
+        gtk_window_set_transient_for (self->saved_networks_dialog,
+                                      GTK_WINDOW (gtk_widget_get_native (GTK_WIDGET (self))));
+        gtk_window_present (self->saved_networks_dialog);
+}
+
+static void
+on_connect_hidden_network (NetDeviceWifi *self)
+{
+        connect_hidden (self);
+}
+
+static void
+on_wifi_hotspot_network (NetDeviceWifi *self)
+{
+        start_hotspot (self);
+}
+
 static void
 ap_activated (NetDeviceWifi *self, GtkListBoxRow *row)
 {
@@ -1175,6 +1072,7 @@ net_device_wifi_class_init (NetDeviceWifiClass *klass)
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
         GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 
+        object_class->dispose = net_device_wifi_dispose;
         object_class->finalize = net_device_wifi_finalize;
         object_class->get_property = net_device_wifi_get_property;
 
@@ -1188,23 +1086,26 @@ net_device_wifi_class_init (NetDeviceWifiClass *klass)
 
         gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/network/network-wifi.ui");
 
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, center_box);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, connect_hidden_button);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, device_off_switch);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, header_box);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, wifi_headerbar_title);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, device_enable_row);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, saved_network_row);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, connect_hidden_row);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, hotspot_row);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, details_box);
         gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, hotspot_box);
         gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, hotspot_name_row);
         gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, hotspot_security_row);
         gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, hotspot_password_row);
         gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, listbox_box);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, start_hotspot_button);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, status_label);
-        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, title_label);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, stack);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, saved_networks_dialog);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, saved_networks_toast_overlay);
+        gtk_widget_class_bind_template_child (widget_class, NetDeviceWifi, saved_networks_box);
 
-        gtk_widget_class_bind_template_callback (widget_class, connect_hidden_button_clicked_cb);
         gtk_widget_class_bind_template_callback (widget_class, device_off_switch_changed_cb);
-        gtk_widget_class_bind_template_callback (widget_class, history_button_clicked_cb);
-        gtk_widget_class_bind_template_callback (widget_class, start_hotspot_button_clicked_cb);
+        gtk_widget_class_bind_template_callback (widget_class, on_show_saved_network);
+        gtk_widget_class_bind_template_callback (widget_class, on_connect_hidden_network);
+        gtk_widget_class_bind_template_callback (widget_class, on_wifi_hotspot_network);
 }
 
 static void
@@ -1231,21 +1132,22 @@ nm_client_on_permission_change (NetDeviceWifi *self) {
         caps = nm_device_wifi_get_capabilities (NM_DEVICE_WIFI (self->device));
         if (perm != NM_CLIENT_PERMISSION_RESULT_YES &&
                 perm != NM_CLIENT_PERMISSION_RESULT_AUTH) {
-                gtk_widget_set_tooltip_text (GTK_WIDGET (self->start_hotspot_button), _("System policy prohibits use as a Hotspot"));
-                gtk_widget_set_sensitive (GTK_WIDGET (self->start_hotspot_button), FALSE);
+                gtk_widget_set_tooltip_text (GTK_WIDGET (self->hotspot_row), _("System policy prohibits use as a Hotspot"));
+                gtk_widget_set_sensitive (GTK_WIDGET (self->hotspot_row), FALSE);
         } else if (!(caps & (NM_WIFI_DEVICE_CAP_AP | NM_WIFI_DEVICE_CAP_ADHOC))) {
-                gtk_widget_set_tooltip_text (GTK_WIDGET (self->start_hotspot_button), _("Wireless device does not support Hotspot mode"));
-                gtk_widget_set_sensitive (GTK_WIDGET (self->start_hotspot_button), FALSE);
-        } else
-                gtk_widget_set_sensitive (GTK_WIDGET (self->start_hotspot_button), TRUE);
-
+                gtk_widget_set_tooltip_text (GTK_WIDGET (self->hotspot_row), _("Wireless device does not support Hotspot mode"));
+                gtk_widget_set_sensitive (GTK_WIDGET (self->hotspot_row), FALSE);
+        } else {
+                gtk_widget_set_sensitive (GTK_WIDGET (self->hotspot_row), nm_client_wireless_get_enabled (self->client));
+        }
 }
 
 NetDeviceWifi *
 net_device_wifi_new (CcPanel *panel, NMClient *client, NMDevice *device)
 {
         NetDeviceWifi *self;
-        GtkWidget *list;
+        GtkListBox *listbox;
+        CcWifiConnectionList *list;
 
         self = g_object_new (net_device_wifi_get_type (), NULL);
         self->panel = panel;
@@ -1257,20 +1159,37 @@ net_device_wifi_new (CcPanel *panel, NMClient *client, NMDevice *device)
 
         g_signal_connect_object (device, "state-changed", G_CALLBACK (nm_device_wifi_refresh_ui), self, G_CONNECT_SWAPPED);
 
-        list = GTK_WIDGET (cc_wifi_connection_list_new (client, NM_DEVICE_WIFI (device), TRUE, TRUE, FALSE));
-        gtk_widget_show (list);
-        gtk_container_add (GTK_CONTAINER (self->listbox_box), list);
+        /* Set up the main Visible Networks list */
+        list = cc_wifi_connection_list_new (client, NM_DEVICE_WIFI (device), TRUE, TRUE, FALSE, FALSE);
+        gtk_box_append (self->listbox_box, GTK_WIDGET (list));
 
-        gtk_list_box_set_header_func (GTK_LIST_BOX (list), cc_list_box_update_header_func, NULL, NULL);
-        gtk_list_box_set_sort_func (GTK_LIST_BOX (list), (GtkListBoxSortFunc)ap_sort, self, NULL);
+        listbox = cc_wifi_connection_list_get_list_box (list);
+        gtk_list_box_set_sort_func (listbox, (GtkListBoxSortFunc)ap_sort, self, NULL);
 
-        g_signal_connect_object (list, "row-activated",
+        g_signal_connect_object (listbox, "row-activated",
                                  G_CALLBACK (ap_activated), self, G_CONNECT_SWAPPED);
         g_signal_connect_object (list, "configure",
                                  G_CALLBACK (show_details_for_row), self, G_CONNECT_SWAPPED);
+        g_signal_connect_object (list, "show-qr-code",
+                                 G_CALLBACK (show_qr_code_for_row), self, G_CONNECT_SWAPPED);
         g_signal_connect_object (client, "notify",
                                  G_CALLBACK(nm_client_on_permission_change), self, G_CONNECT_SWAPPED);
 
+        /* Set up the Saved Networks list */
+        list = cc_wifi_connection_list_new (self->client, NM_DEVICE_WIFI (device), FALSE, FALSE, FALSE, TRUE);
+        self->saved_networks_list = g_object_ref_sink (list);
+        adw_preferences_group_add (self->saved_networks_box, GTK_WIDGET (list));
+
+        listbox = cc_wifi_connection_list_get_list_box (list);
+        gtk_list_box_set_sort_func (listbox, (GtkListBoxSortFunc)history_sort, NULL, NULL);
+
+        g_signal_connect_object (list, "configure",
+                                 G_CALLBACK (show_details_for_row), self, G_CONNECT_SWAPPED);
+        g_signal_connect_object (list, "forget",
+                                 G_CALLBACK (forget_selected), self, G_CONNECT_SWAPPED);
+        g_signal_connect_object (list, "show_qr_code",
+                                 G_CALLBACK (show_qr_code_for_row), self, G_CONNECT_SWAPPED);
+
         nm_client_on_permission_change(self);
 
         nm_device_wifi_refresh_ui (self);
@@ -1289,21 +1208,21 @@ void
 net_device_wifi_set_title (NetDeviceWifi *self, const gchar *title)
 {
         g_return_if_fail (NET_IS_DEVICE_WIFI (self));
-        gtk_label_set_label (self->title_label, title);
+        adw_window_title_set_title (self->wifi_headerbar_title, title);
 }
 
 GtkWidget *
 net_device_wifi_get_header_widget (NetDeviceWifi *self)
 {
         g_return_val_if_fail (NET_IS_DEVICE_WIFI (self), NULL);
-        return GTK_WIDGET (self->header_box);
+        return GTK_WIDGET (self->details_box);
 }
 
 GtkWidget *
 net_device_wifi_get_title_widget (NetDeviceWifi *self)
 {
         g_return_val_if_fail (NET_IS_DEVICE_WIFI (self), NULL);
-        return GTK_WIDGET (self->center_box);
+        return GTK_WIDGET (self->wifi_headerbar_title);
 }
 
 void
