@@ -238,8 +238,7 @@ row_data_new (CcPanelCategory     category,
               const gchar        *description,
               const GStrv         keywords,
               const gchar        *icon,
-              CcPanelVisibility   visibility,
-              gboolean            has_sidebar)
+              CcPanelVisibility   visibility)
 {
   GtkWidget *label, *grid, *image;
   RowData *data;
@@ -290,12 +289,6 @@ row_data_new (CcPanelCategory     category,
                                   NULL,
                                   -1);
 
-  if (has_sidebar)
-    {
-      image = gtk_image_new_from_icon_name ("go-next-symbolic");
-      gtk_grid_attach (GTK_GRID (grid), image, 2, 0, 1, 1);
-    }
-
   gtk_widget_add_css_class (label, "dim-label");
   gtk_grid_attach (GTK_GRID (grid), label, 1, 1, 1, 1);
 
@@ -361,35 +354,37 @@ static const gchar * const panel_order[] = {
   "wwan",
   "mobile-broadband",
   "bluetooth",
+
+  "separator",
+
+  "display",
+  "sound",
+  "power",
+  "multitasking",
   "background",
-  "ubuntu",
+
+  "separator",
+
+  "applications",
   "notifications",
   "search",
-  "multitasking",
-  "applications",
-  "privacy",
   "online-accounts",
   "sharing",
 
-  /* Devices page */
-  "sound",
-  "power",
-  "display",
+  "separator",
+
   "mouse",
   "keyboard",
-  "printers",
-  "removable-media",
-  "wacom",
   "color",
+  "printers",
+  "wacom",
+
+  "separator",
 
-  /* Details page */
-  "region",
   "universal-access",
-  "user-accounts",
-  "default-apps",
+  "privacy",
+  "system",
   "reset-settings",
-  "datetime",
-  "info-overview",
 };
 
 static guint
@@ -472,15 +467,13 @@ header_func (GtkListBoxRow *row,
              GtkListBoxRow *before,
              gpointer       user_data)
 {
-  RowData *row_data, *before_data;
+  guint pid;
 
   if (!before)
     return;
 
-  row_data = g_object_get_data (G_OBJECT (row), "data");
-  before_data = g_object_get_data (G_OBJECT (before), "data");
-
-  if (row_data->category != before_data->category)
+  pid = get_panel_id_index (get_panel_id_from_row (user_data, row));
+  if (g_str_equal (panel_order[pid-1], "separator"))
     {
       GtkWidget *separator;
 
@@ -723,7 +716,7 @@ cc_panel_list_class_init (CcPanelListClass *klass)
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
 
-  gtk_widget_class_set_template_from_resource (widget_class, "/org/regolith/Settings/gtk/cc-panel-list.ui");
+  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Settings/gtk/cc-panel-list.ui");
 
   gtk_widget_class_bind_template_child (widget_class, CcPanelList, main_listbox);
   gtk_widget_class_bind_template_child (widget_class, CcPanelList, search_listbox);
@@ -872,21 +865,20 @@ cc_panel_list_add_panel (CcPanelList        *self,
                          const gchar        *description,
                          const GStrv         keywords,
                          const gchar        *icon,
-                         CcPanelVisibility   visibility,
-                         gboolean            has_sidebar)
+                         CcPanelVisibility   visibility)
 {
   RowData *data, *search_data;
 
   g_return_if_fail (CC_IS_PANEL_LIST (self));
 
   /* Add the panel to the proper listbox */
-  data = row_data_new (category, id, title, description, keywords, icon, visibility, has_sidebar);
+  data = row_data_new (category, id, title, description, keywords, icon, visibility);
   gtk_widget_set_visible (data->row, visibility == CC_PANEL_VISIBLE);
 
   gtk_list_box_append (GTK_LIST_BOX (self->main_listbox), data->row);
 
   /* And add to the search listbox too */
-  search_data = row_data_new (category, id, title, description, keywords, icon, visibility, has_sidebar);
+  search_data = row_data_new (category, id, title, description, keywords, icon, visibility);
   gtk_widget_set_visible (search_data->row, visibility != CC_PANEL_HIDDEN);
 
   gtk_list_box_append (GTK_LIST_BOX (self->search_listbox), search_data->row);
